/* md5.c - Functions to compute MD5 message digest of files or memory blocks
   according to the definition of MD5 in RFC 1321 from April 1992.
   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* Written by Ulrich Drepper <drepper@gnu.ai.mit.edu>, 1995.  */

#include "config.h"

#include <sys/types.h>

#include <stdlib.h>
#include <string.h>
#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif

#ifdef HAVE_OPENSSL
#include <openssl/opensslv.h>
#include <openssl/md5.h>

#if OPENSSL_VERSION_NUMBER >= 0x10101000
#define HAVE_OPENSSL_EVP 1
#include <openssl/evp.h>
#endif

#endif /* HAVE_OPENSSL */

#include "ne_md5.h"
#include "ne_string.h" /* for NE_ASC2HEX */

#if SIZEOF_INT == 4
typedef unsigned int md5_uint32;
#elif SIZEOF_LONG == 4
typedef unsigned long md5_uint32;
#else
# error "Cannot determine unsigned 32-bit data type."
#endif

#define md5_process_block ne_md5_process_block
#define md5_process_bytes ne_md5_process_bytes
#define md5_finish_ctx ne_md5_finish_ctx
#define md5_read_ctx ne_md5_read_ctx
#define md5_stream ne_md5_stream
#define md5_ctx ne_md5_ctx


#ifdef WORDS_BIGENDIAN
# define SWAP(n)							\
    (((n) << 24) | (((n) & 0xff00) << 8) | (((n) >> 8) & 0xff00) | ((n) >> 24))
#else
# define SWAP(n) (n)
#endif

/* Structure to save state of computation between the single steps.  */
struct md5_ctx
{
#ifdef HAVE_OPENSSL_EVP
  EVP_MD_CTX *ctx;
#elif defined(HAVE_OPENSSL)
  MD5_CTX ctx;
#else
  md5_uint32 A;
  md5_uint32 B;
  md5_uint32 C;
  md5_uint32 D;

  md5_uint32 total[2];
  md5_uint32 buflen;
  char buffer[128];
#endif
};

#ifndef HAVE_OPENSSL
/* This array contains the bytes used to pad the buffer to the next
   64-byte boundary.  (RFC 1321, 3.1: Step 1)  */
static const unsigned char fillbuf[64] = { 0x80, 0 /* , 0, 0, ...  */ };


/* Initialize structure containing state of computation.
   (RFC 1321, 3.3: Step 3)  */
static void 
md5_init_ctx (struct md5_ctx *ctx)
{
  ctx->A = 0x67452301;
  ctx->B = 0xefcdab89;
  ctx->C = 0x98badcfe;
  ctx->D = 0x10325476;

  ctx->total[0] = ctx->total[1] = 0;
  ctx->buflen = 0;
}

struct ne_md5_ctx *
ne_md5_create_ctx(void)
{
  struct md5_ctx *ctx = ne_malloc(sizeof *ctx);
  md5_init_ctx(ctx);
  return ctx;
}

extern void 
ne_md5_reset_ctx(struct ne_md5_ctx *ctx)
{
  md5_init_ctx(ctx);
}

struct ne_md5_ctx *
ne_md5_dup_ctx(struct ne_md5_ctx *ctx)
{
  return memcpy(ne_malloc(sizeof *ctx), ctx, sizeof *ctx);
}

void
ne_md5_destroy_ctx(struct ne_md5_ctx *ctx)
{
  ne_free(ctx);
}

/* Process the remaining bytes in the internal buffer and the usual
   prolog according to the standard and write the result to RESBUF.

   IMPORTANT: On some systems it is required that RESBUF is correctly
   aligned for a 32 bits value.  */
void *
md5_finish_ctx (struct md5_ctx *ctx, void *resbuf)
{
  /* Take yet unprocessed bytes into account.  */
  md5_uint32 bytes = ctx->buflen;
  md5_uint32 swap_bytes;
  size_t pad;

  /* Now count remaining bytes.  */
  ctx->total[0] += bytes;
  if (ctx->total[0] < bytes)
    ++ctx->total[1];

  pad = bytes >= 56 ? 64 + 56 - bytes : 56 - bytes;
  memcpy (&ctx->buffer[bytes], fillbuf, pad);

  /* Put the 64-bit file length in *bits* at the end of the buffer.
     Use memcpy to avoid aliasing problems.  On most systems, this
     will be optimized away to the same code.  */
  swap_bytes = SWAP (ctx->total[0] << 3);
  memcpy (&ctx->buffer[bytes + pad], &swap_bytes, sizeof (swap_bytes));
  swap_bytes = SWAP ((ctx->total[1] << 3) | (ctx->total[0] >> 29));
  memcpy (&ctx->buffer[bytes + pad + 4], &swap_bytes, sizeof (swap_bytes));

  /* Process last bytes.  */
  md5_process_block (ctx->buffer, bytes + pad + 8, ctx);

  return md5_read_ctx (ctx, resbuf);
}

void
md5_process_bytes (const void *buffer, size_t len, struct md5_ctx *ctx)
{
  /* When we already have some bits in our internal buffer concatenate
     both inputs first.  */
  if (ctx->buflen != 0)
    {
      size_t left_over = ctx->buflen;
      size_t add = 128 - left_over > len ? len : 128 - left_over;

      memcpy (&ctx->buffer[left_over], buffer, add);
      ctx->buflen += add;

      if (left_over + add > 64)
	{
	  md5_process_block (ctx->buffer, (left_over + add) & ~63, ctx);
	  /* The regions in the following copy operation cannot overlap.  */
	  memcpy (ctx->buffer, &ctx->buffer[(left_over + add) & ~63],
		  (left_over + add) & 63);
	  ctx->buflen = (left_over + add) & 63;
	}

      buffer = (const char *) buffer + add;
      len -= add;
    }

  /* Process available complete blocks.  */
  if (len > 64)
    {
      md5_process_block (buffer, len & ~63, ctx);
      buffer = (const char *) buffer + (len & ~63);
      len &= 63;
    }

  /* Move remaining bytes in internal buffer.  */
  if (len > 0)
    {
      memcpy (ctx->buffer, buffer, len);
      ctx->buflen = len;
    }
}


/* These are the four functions used in the four steps of the MD5 algorithm
   and defined in the RFC 1321.  The first function is a little bit optimized
   (as found in Colin Plumbs public domain implementation).  */
/* #define FF(b, c, d) ((b & c) | (~b & d)) */
#define FF(b, c, d) (d ^ (b & (c ^ d)))
#define FG(b, c, d) FF (d, b, c)
#define FH(b, c, d) (b ^ c ^ d)
#define FI(b, c, d) (c ^ (b | ~d))

/* Process LEN bytes of BUFFER, accumulating context into CTX.
   It is assumed that LEN % 64 == 0.  */

void
md5_process_block (const void *buffer, size_t len, struct md5_ctx *ctx)
{
  md5_uint32 correct_words[16];
  const unsigned char *words = buffer;
  const unsigned char *endp = words + len;
  md5_uint32 A = ctx->A;
  md5_uint32 B = ctx->B;
  md5_uint32 C = ctx->C;
  md5_uint32 D = ctx->D;

  /* First increment the byte count.  RFC 1321 specifies the possible
     length of the file up to 2^64 bits.  Here we only compute the
     number of bytes.  Do a double word increment.  */
  ctx->total[0] += len;
  if (ctx->total[0] < len)
    ++ctx->total[1];

  /* Process all bytes in the buffer with 64 bytes in each round of
     the loop.  */
  while (words < endp)
    {
      md5_uint32 *cwp = correct_words;
      md5_uint32 A_save = A;
      md5_uint32 B_save = B;
      md5_uint32 C_save = C;
      md5_uint32 D_save = D;

      /* First round: using the given function, the context and a constant
	 the next context is computed.  Because the algorithms processing
	 unit is a 32-bit word and it is determined to work on words in
	 little endian byte order we perhaps have to change the byte order
	 before the computation.  To reduce the work for the next steps
	 we store the swapped words in the array CORRECT_WORDS.  */

#define OP(a, b, c, d, s, T)						\
      do								\
        {								\
	  md5_uint32 WORD_ = (md5_uint32)words[0] | ((md5_uint32)words[1] << 8) \
	       | ((md5_uint32)words[2] << 16) | ((md5_uint32)words[3] << 24); \
	  a += FF (b, c, d) + (*cwp++ = WORD_) + T;		\
	  words += 4;							\
	  CYCLIC (a, s);						\
	  a += b;							\
        }								\
      while (0)

      /* It is unfortunate that C does not provide an operator for
	 cyclic rotation.  Hope the C compiler is smart enough.  */
#define CYCLIC(w, s) (w = (w << s) | (w >> (32 - s)))

      /* Before we start, one word to the strange constants.
	 They are defined in RFC 1321 as

	 T[i] = (int) (4294967296.0 * fabs (sin (i))), i=1..64
       */

      /* Round 1.  */
      OP (A, B, C, D,  7, 0xd76aa478);
      OP (D, A, B, C, 12, 0xe8c7b756);
      OP (C, D, A, B, 17, 0x242070db);
      OP (B, C, D, A, 22, 0xc1bdceee);
      OP (A, B, C, D,  7, 0xf57c0faf);
      OP (D, A, B, C, 12, 0x4787c62a);
      OP (C, D, A, B, 17, 0xa8304613);
      OP (B, C, D, A, 22, 0xfd469501);
      OP (A, B, C, D,  7, 0x698098d8);
      OP (D, A, B, C, 12, 0x8b44f7af);
      OP (C, D, A, B, 17, 0xffff5bb1);
      OP (B, C, D, A, 22, 0x895cd7be);
      OP (A, B, C, D,  7, 0x6b901122);
      OP (D, A, B, C, 12, 0xfd987193);
      OP (C, D, A, B, 17, 0xa679438e);
      OP (B, C, D, A, 22, 0x49b40821);

      /* For the second to fourth round we have the possibly swapped words
	 in CORRECT_WORDS.  Redefine the macro to take an additional first
	 argument specifying the function to use.  */
#undef OP
#define OP(f, a, b, c, d, k, s, T)					\
      do 								\
	{								\
	  a += f (b, c, d) + correct_words[k] + T;			\
	  CYCLIC (a, s);						\
	  a += b;							\
	}								\
      while (0)

      /* Round 2.  */
      OP (FG, A, B, C, D,  1,  5, 0xf61e2562);
      OP (FG, D, A, B, C,  6,  9, 0xc040b340);
      OP (FG, C, D, A, B, 11, 14, 0x265e5a51);
      OP (FG, B, C, D, A,  0, 20, 0xe9b6c7aa);
      OP (FG, A, B, C, D,  5,  5, 0xd62f105d);
      OP (FG, D, A, B, C, 10,  9, 0x02441453);
      OP (FG, C, D, A, B, 15, 14, 0xd8a1e681);
      OP (FG, B, C, D, A,  4, 20, 0xe7d3fbc8);
      OP (FG, A, B, C, D,  9,  5, 0x21e1cde6);
      OP (FG, D, A, B, C, 14,  9, 0xc33707d6);
      OP (FG, C, D, A, B,  3, 14, 0xf4d50d87);
      OP (FG, B, C, D, A,  8, 20, 0x455a14ed);
      OP (FG, A, B, C, D, 13,  5, 0xa9e3e905);
      OP (FG, D, A, B, C,  2,  9, 0xfcefa3f8);
      OP (FG, C, D, A, B,  7, 14, 0x676f02d9);
      OP (FG, B, C, D, A, 12, 20, 0x8d2a4c8a);

      /* Round 3.  */
      OP (FH, A, B, C, D,  5,  4, 0xfffa3942);
      OP (FH, D, A, B, C,  8, 11, 0x8771f681);
      OP (FH, C, D, A, B, 11, 16, 0x6d9d6122);
      OP (FH, B, C, D, A, 14, 23, 0xfde5380c);
      OP (FH, A, B, C, D,  1,  4, 0xa4beea44);
      OP (FH, D, A, B, C,  4, 11, 0x4bdecfa9);
      OP (FH, C, D, A, B,  7, 16, 0xf6bb4b60);
      OP (FH, B, C, D, A, 10, 23, 0xbebfbc70);
      OP (FH, A, B, C, D, 13,  4, 0x289b7ec6);
      OP (FH, D, A, B, C,  0, 11, 0xeaa127fa);
      OP (FH, C, D, A, B,  3, 16, 0xd4ef3085);
      OP (FH, B, C, D, A,  6, 23, 0x04881d05);
      OP (FH, A, B, C, D,  9,  4, 0xd9d4d039);
      OP (FH, D, A, B, C, 12, 11, 0xe6db99e5);
      OP (FH, C, D, A, B, 15, 16, 0x1fa27cf8);
      OP (FH, B, C, D, A,  2, 23, 0xc4ac5665);

      /* Round 4.  */
      OP (FI, A, B, C, D,  0,  6, 0xf4292244);
      OP (FI, D, A, B, C,  7, 10, 0x432aff97);
      OP (FI, C, D, A, B, 14, 15, 0xab9423a7);
      OP (FI, B, C, D, A,  5, 21, 0xfc93a039);
      OP (FI, A, B, C, D, 12,  6, 0x655b59c3);
      OP (FI, D, A, B, C,  3, 10, 0x8f0ccc92);
      OP (FI, C, D, A, B, 10, 15, 0xffeff47d);
      OP (FI, B, C, D, A,  1, 21, 0x85845dd1);
      OP (FI, A, B, C, D,  8,  6, 0x6fa87e4f);
      OP (FI, D, A, B, C, 15, 10, 0xfe2ce6e0);
      OP (FI, C, D, A, B,  6, 15, 0xa3014314);
      OP (FI, B, C, D, A, 13, 21, 0x4e0811a1);
      OP (FI, A, B, C, D,  4,  6, 0xf7537e82);
      OP (FI, D, A, B, C, 11, 10, 0xbd3af235);
      OP (FI, C, D, A, B,  2, 15, 0x2ad7d2bb);
      OP (FI, B, C, D, A,  9, 21, 0xeb86d391);

      /* Add the starting values of the context.  */
      A += A_save;
      B += B_save;
      C += C_save;
      D += D_save;
    }

  /* Put checksum in context given as argument.  */
  ctx->A = A;
  ctx->B = B;
  ctx->C = C;
  ctx->D = D;
}
#elif defined(HAVE_OPENSSL_EVP)

struct ne_md5_ctx *ne_md5_create_ctx(void)
{
    struct ne_md5_ctx *ret;
    EVP_MD_CTX *ctx;

    ctx = EVP_MD_CTX_new();
    if (!ctx) return NULL;

    if (EVP_DigestInit(ctx, EVP_md5()) != 1) {
        EVP_MD_CTX_free(ctx);
        return NULL;
    }

    ret = ne_malloc(sizeof *ret);
    ret->ctx = ctx;
    return ret;
}

void ne_md5_process_block(const void *buffer, size_t len,
                          struct ne_md5_ctx *ctx)
{
    EVP_DigestUpdate(ctx->ctx, buffer, len);
}

void ne_md5_process_bytes(const void *buffer, size_t len,
                          struct ne_md5_ctx *ctx)
{
    EVP_DigestUpdate(ctx->ctx, buffer, len);
}

void *ne_md5_finish_ctx(struct ne_md5_ctx *ctx, void *resbuf)
{
    unsigned int len = MD5_DIGEST_LENGTH;
    EVP_DigestFinal(ctx->ctx, resbuf, &len);
    return resbuf;
}

struct ne_md5_ctx *ne_md5_dup_ctx(struct ne_md5_ctx *ctx)
{
    struct ne_md5_ctx *ret;
    EVP_MD_CTX *newctx;

    newctx = EVP_MD_CTX_new();
    if (!newctx) return NULL;

    if (EVP_MD_CTX_copy_ex(newctx, ctx->ctx) != 1) {
        EVP_MD_CTX_free(newctx);
        return NULL;
    }

    ret = ne_malloc(sizeof *ret);
    ret->ctx = newctx;
    return ret;
}

void ne_md5_reset_ctx(struct ne_md5_ctx *ctx)
{
    EVP_MD_CTX_reset(ctx->ctx);
}

void ne_md5_destroy_ctx(struct ne_md5_ctx *ctx)
{
    EVP_MD_CTX_free(ctx->ctx);
    ne_free(ctx);
}

#else /* OpenSSL < 3.1 */
struct ne_md5_ctx *ne_md5_create_ctx(void)
{
    struct ne_md5_ctx *ctx = ne_malloc(sizeof *ctx);
    
    if (MD5_Init(&ctx->ctx) != 1) {
        ne_free(ctx);
        return NULL;
    }
    
    return ctx;
}

void ne_md5_process_block(const void *buffer, size_t len,
                          struct ne_md5_ctx *ctx)
{
    MD5_Update(&ctx->ctx, buffer, len);
}

void ne_md5_process_bytes(const void *buffer, size_t len,
                          struct ne_md5_ctx *ctx)
{
    MD5_Update(&ctx->ctx, buffer, len);
}

void *ne_md5_finish_ctx(struct ne_md5_ctx *ctx, void *resbuf)
{
    MD5_Final(resbuf, &ctx->ctx);
    
    return resbuf;
}

struct ne_md5_ctx *ne_md5_dup_ctx(struct ne_md5_ctx *ctx)
{
    return memcpy(ne_malloc(sizeof *ctx), ctx, sizeof *ctx);
}

void ne_md5_reset_ctx(struct ne_md5_ctx *ctx)
{
    MD5_Init(&ctx->ctx);
}
    
void ne_md5_destroy_ctx(struct ne_md5_ctx *ctx)
{
    ne_free(ctx);
}
#endif /* HAVE_OPENSSL */

/* Put result from CTX in first 16 bytes following RESBUF.  The result
   must be in little endian byte order.

   IMPORTANT: On some systems it is required that RESBUF is correctly
   aligned for a 32 bits value.  */
void *
md5_read_ctx (const struct md5_ctx *ctx, void *resbuf)
{
#ifdef HAVE_OPENSSL_EVP
  return NULL;
#else

#ifdef HAVE_OPENSSL
#define SWAP_CTX(x) SWAP(ctx->ctx.x)
#else
#define SWAP_CTX(x) SWAP(ctx->x)
#endif

  ((md5_uint32 *) resbuf)[0] = SWAP_CTX (A);
  ((md5_uint32 *) resbuf)[1] = SWAP_CTX (B);
  ((md5_uint32 *) resbuf)[2] = SWAP_CTX (C);
  ((md5_uint32 *) resbuf)[3] = SWAP_CTX (D);

  return resbuf;
#endif /* !HAVE_OPENSSL_EVP */
}


/* Compute MD5 message digest for bytes read from STREAM.  The
   resulting message digest number will be written into the 16 bytes
   beginning at RESBLOCK.  */
int
md5_stream (FILE *stream, void *resblock)
{
  /* Important: BLOCKSIZE must be a multiple of 64.  */
#define BLOCKSIZE 4096
  struct ne_md5_ctx *ctx;
  char buffer[BLOCKSIZE + 72];
  size_t sum;

  /* Initialize the computation context.  */
  ctx = ne_md5_create_ctx ();

  /* Iterate over full file contents.  */
  while (1)
    {
      /* We read the file in blocks of BLOCKSIZE bytes.  One call of the
	 computation function processes the whole buffer so that with the
	 next round of the loop another block can be read.  */
      size_t n;
      sum = 0;

      /* Read block.  Take care for partial reads.  */
      do
	{
	  n = fread (buffer + sum, 1, BLOCKSIZE - sum, stream);

	  sum += n;
	}
      while (sum < BLOCKSIZE && n != 0);
      if (n == 0 && ferror (stream))
        return 1;

      /* If end of file is reached, end the loop.  */
      if (n == 0)
	break;

      /* Process buffer with BLOCKSIZE bytes.  Note that
			BLOCKSIZE % 64 == 0
       */
      md5_process_block (buffer, BLOCKSIZE, ctx);
    }

  /* Add the last bytes if necessary.  */
  if (sum > 0)
    md5_process_bytes (buffer, sum, ctx);

  /* Construct result in desired memory.  */
  md5_finish_ctx (ctx, resblock);
  ne_md5_destroy_ctx (ctx);
  
  return 0;
}

/* Writes the ASCII representation of the MD5 digest into the
 * given buffer, which must be at least 33 characters long. */
void ne_md5_to_ascii(const unsigned char md5_buf[16], char *buffer) 
{
    int count;
    for (count = 0; count<16; count++) {
	buffer[count*2] = NE_HEX2ASC(md5_buf[count] >> 4);
	buffer[count*2+1] = NE_HEX2ASC(md5_buf[count] & 0x0f);
    }
    buffer[32] = '\0';
}

/* Reads the ASCII representation of an MD5 digest. The buffer must
 * be at least 32 characters long. */
void ne_ascii_to_md5(const char *buffer, unsigned char md5_buf[16]) 
{
    int count;
    for (count = 0; count<16; count++) {
	md5_buf[count] = ((NE_ASC2HEX(buffer[count*2])) << 4) |
	    NE_ASC2HEX(buffer[count*2+1]);
    }
}

char *ne_md5_finish_ascii(struct ne_md5_ctx *ctx, char buffer[33])
{
    md5_uint32 result[4];

    ne_md5_finish_ctx(ctx, (void *)result);
    ne_md5_to_ascii((void *)result, buffer);

    return buffer;
}

